from ..scaler import StandardScaler, Scaler, MaxAbsScaler
from .sliding_window import SlidingWindow
from .sliding_window_ts import SlidingWindowTS
from .maskts import MaskTimeFeatureSet, MaskTS
from .uea import UEAClassification
from .anomaly import AnomalyLoader


forecast_loaders = [
    "SlidingWindowTS",
    "SlidingWindow",
]

imputation_loaders = [
    "MaskTS",
    "MaskTimeFeatureSet",
]


classification_loaders = [
    "UEAClassification",
]

anomaly_loaders = [
    "AnomalyLoader",
]

scalers = [
    "StandardScalar",
    "MaxAbsScaler",
    "MinMaxScaler",
]

data_loaders = (
    forecast_loaders + anomaly_loaders + classification_loaders + imputation_loaders
)
__all__ = data_loaders + scalers



from sktime.datasets import load_from_tsfile_to_dataframe


import os
import resource
from .dataset import Dataset, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import (
    download_url,
    download_and_extract_archive,
    check_integrity,
)
import pandas as pd
import numpy as np
import torch.utils.data

def subsample(y, limit=256, factor=2):
    """
    If a given Series is longer than `limit`, returns subsampled sequence by the specified integer factor
    """
    if len(y) > limit:
        return y[::factor].reset_index(drop=True)
    return y


def interpolate_missing(y):
    """
    Replaces NaN values in pd.Series `y` using linear interpolation
    """
    if y.isna().any():
        y = y.interpolate(method='linear', limit_direction='both')
    return y


class UEA(TimeSeriesDataset):
    """
    Dataset class for datasets included in:
    Time Series Classification Archive (www.timeseriesclassification.com).
    
    For a list of available datasets, see https://www.timeseriesclassification.com/dataset.php.

    Attributes:
        train_df (pd.DataFrame): DataFrame containing the training data.
        train_labels (pd.DataFrame): DataFrame containing the labels for the training data.
        test_df (pd.DataFrame): DataFrame containing the test data.
        test_labels (pd.DataFrame): DataFrame containing the labels for the test data.
        num_classes (int): Number of classes in the dataset.
    """
    
    train_df : pd.DataFrame
    train_labels  : pd.DataFrame
    test_df : pd.DataFrame
    test_labels  : pd.DataFrame
    num_classes  : int

    def __init__(self, name, root='./data'):
        self.name = name
        self.root = root
        self.dir = os.path.join(root, self.name)
        os.makedirs(self.dir, exist_ok=True)
        
        self.download()
        
        self._load()
        
        self.num_features = self.train_features_data.shape[1]
        
    def download(self):
        download_and_extract_archive(
            f"https://www.timeseriesclassification.com/aeon-toolkit/{self.name}.zip",
            self.dir,
            filename=f"{self.name}.zip",
        )


    def load_single(self, filepath):
        df, labels = load_from_tsfile_to_dataframe(filepath, return_separate_X_and_y=True,
                                                             replace_missing_vals_with='NaN')
        labels = pd.Series(labels, dtype="category")
        self.class_names = labels.cat.categories
        labels_df = pd.DataFrame(labels.cat.codes,
                                 dtype=np.int8)  # int8-32 gives an error when using nn.CrossEntropyLoss

        lengths = df.applymap(
            lambda x: len(x)).values  # (num_samples, num_dimensions) array containing the length of each series

        horiz_diffs = np.abs(lengths - np.expand_dims(lengths[:, 0], -1))

        if np.sum(horiz_diffs) > 0:  # if any row (sample) has varying length across dimensions
            df = df.applymap(subsample)

        lengths = df.applymap(lambda x: len(x)).values
        vert_diffs = np.abs(lengths - np.expand_dims(lengths[0, :], 0))
        if np.sum(vert_diffs) > 0:  # if any column (dimension) has varying length across samples
            self.max_seq_len = int(np.max(lengths[:, 0]))
        else:
            self.max_seq_len = lengths[0, 0]

        # First create a (seq_len, feat_dim) dataframe for each sample, indexed by a single integer ("ID" of the sample)
        # Then concatenate into a (num_samples * seq_len, feat_dim) dataframe, with multiple rows corresponding to the
        # sample index (i.e. the same scheme as all datasets in this project)

        df = pd.concat((pd.DataFrame({col: df.loc[row, col] for col in df.columns}).reset_index(drop=True).set_index(
            pd.Series(lengths[row, 0] * [row])) for row in range(df.shape[0])), axis=0)

        # Replace NaN values
        grp = df.groupby(by=df.index)
        df = grp.transform(interpolate_missing)

        return df, labels_df
    
    
    def _load(self) -> np.ndarray:
        self.train_filepath = os.path.join(self.dir, f"{self.name}_TRAIN.ts")
        self.test_filepath = os.path.join(self.dir, f"{self.name}_TEST.ts")

        self.train_df, self.train_labels = self.load_single(self.train_filepath)
        self.test_df, self.test_labels = self.load_single(self.test_filepath)
        
        self.train_features_data, self.train_labels_data = self.train_df.values, self.train_labels.values
        self.test_features_data, self.test_labels_data = self.test_df.values, self.test_labels.values

        self.num_classes = len(self.class_names)
                
        
        
        
        



from .dummies import Dummy, DummyGraph
from .dataset import TimeSeriesDataset, TimeseriesSubset, AnomalyDataset
from .Traffic import Traffic
from .ExchangeRate import ExchangeRate
from .SolarEnergy import SolarEnergy
from .Electricity import Electricity
from .ETTh1 import ETTh1
from .ETTh2 import ETTh2
from .ETTm1 import ETTm1
from .ILI import ILI
from .Weather import Weather
from .ETTm2 import ETTm2
from .M4 import M4
from .UEA import UEA
from .SWaT import SWaT
from .SMD import SMD
from .MSL import MSL
from .PSM import PSM
from .SMAP import SMAP


forecast_datasets = [
    'Traffic',
    'ExchangeRate',
    'SolarEnergy',
    'Electricity',
    'ETTh1',
    'ETTh2',
    'ETTm1',
    'ETTm2',
    'Weather',
    'ILI',
    'M4',
]


classify_datasets = [
    'UEA'
] 


anomaly_datasets = [
    'SWaT',
    'SMD',
    'SMAP',
    'MSL',
    'PSM',
] 


synthetic_datasets = [
    'Dummy',
    'DummyGraph'
]

__all__ = forecast_datasets + classify_datasets + anomaly_datasets + synthetic_datasets


import numpy as np
from ..scaler import Scaler
from torch_timeseries.dataset import TimeSeriesDataset, AnomalyDataset
from torch.utils.data import Dataset, DataLoader, RandomSampler, Subset


class AnomalySliced(TimeSeriesDataset):
    """
    Data loader for anomaly detection in time series datasets.

    Attributes:
        train_ratio (float): Ratio of the dataset to be used for training.
        val_ratio (float): Ratio of the dataset to be used for validation.
        batch_size (int): Number of samples per batch.
        num_worker (int): Number of worker threads for data loading.
        dataset (TimeSeriesDataset): Time series dataset to be used.
        scaler (Scaler): Scaler to normalize the data.
        window (int): Window size for the time series data.
        shuffle_train (bool): Whether to shuffle the training data.
        train_loader (DataLoader): DataLoader for the training data.
        val_loader (DataLoader): DataLoader for the validation data.
    """

    def __init__(
        self,
        dataset: AnomalyDataset,
        scaler: Scaler,
        train_ratio: float = 0.8,
        step=1,
        window: int = 168,
        scaler_fit=True,
        flag="train",
    ):
        self.dataset = dataset
        self.window = window
        self.step = step
        self.scaler = scaler
        self.flag = flag
        self.train_ratio = train_ratio

        _len = self.dataset.train_data.shape[0]

        self.train_len = int(_len * self.train_ratio)
        if flag == "train":
            self.train_data = self.dataset.train_data[: self.train_len]
            if scaler_fit:
                self.scaler.fit(self.train_data)
            self.scaled_train_data = self.scaler.transform(self.train_data)
        elif flag == "val":
            self.val_data = self.dataset.train_data[self.train_len :]
            self.scaled_val_data = self.scaler.transform(self.val_data)
        elif flag == "test":
            self.test_data = self.dataset.test_data
            self.test_labels = self.dataset.test_labels
            self.scaled_test_data = self.scaler.transform(self.test_data)

    def __getitem__(self, index):
        index = index * self.step
        if self.flag == "train":
            return np.float32(
                self.scaled_train_data[index : index + self.window]
            ), np.float32(self.train_data[index : index + self.window])
        elif self.flag == "val":
            return (
                np.float32(self.scaled_val_data[index : index + self.window]),
                np.float32(self.val_data[index : index + self.window]),
            )
        elif self.flag == "test":
            return (
                np.float32(self.scaled_test_data[index : index + self.window]),
                np.float32(self.test_data[index : index + self.window]),
                np.float32(self.test_labels[index : index + self.window]),
            )
        else:
            NotImplementedError("not implemented")

    def __len__(self):
        if self.flag == "train":
            return (self.train_data.shape[0] - self.window) // self.step + 1
        elif self.flag == "val":
            return (self.val_data.shape[0] - self.window) // self.step + 1
        elif self.flag == "test":
            return (self.test_data.shape[0] - self.window) // self.step + 1
        else:
            raise NotImplementedError("Not implemented!!!")


class AnomalyLoader:

    def __init__(
        self,
        dataset: TimeSeriesDataset,
        scaler: Scaler,
        window: int = 168,
        shuffle_train=True,
        batch_size: int = 32,
        train_ratio: float = 0.8,
        num_worker: int = 3,
    ) -> None:

        self.train_ratio = train_ratio
        self.val_ratio = 1 - self.train_ratio
        self.batch_size = batch_size
        self.num_worker = num_worker
        self.dataset = dataset

        self.scaler = scaler
        self.window = window
        self.shuffle_train = shuffle_train
        
        self.train_loader : DataLoader
        self.val_loader : DataLoader
        self.test_loader : DataLoader

        self._load()

    def _load(self):
        self._load_dataset()
        self._load_dataloader()

    def _load_dataset(self):
        self.train_dataset = AnomalySliced(
            self.dataset,
            self.scaler,
            self.train_ratio,
            self.window,
            scaler_fit=True,
            flag="train",
        )
        self.val_dataset = AnomalySliced(
            self.dataset,
            self.scaler,
            self.train_ratio,
            self.window,
            scaler_fit=False,
            flag="val",
        )
        self.test_dataset = AnomalySliced(
            self.dataset,
            self.scaler,
            self.train_ratio,
            self.window,
            scaler_fit=False,
            flag="test",
        )

    def _load_dataloader(self):
        self.train_size = len(self.train_dataset)
        self.val_size = len(self.val_dataset)
        self.test_size = len(self.test_dataset)
        self.train_loader = DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=self.shuffle_train,
            num_workers=self.num_worker,
        )

        self.val_loader = DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_worker,
        )

        self.test_loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_worker,
        )



import os
import resource
from .dataset import Dataset, TimeSeriesDataset, AnomalyDataset
from typing import Callable, List, Optional
import torch
from torchvision.datasets.utils import download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np

class PSM(AnomalyDataset):
    """
    PSM is a public dataset from eBay's internal server nodes. It has 26 dimensions and contains performance indicators and system metrics from multiple servers, such as CPU usage, memory usage, network traffic, disk IO, etc.
    """
    
    name:str= 'PSM'
    
    def download(self) -> None:
        download_and_extract_archive(
            f"https://drive.usercontent.google.com/download?id=14gCVQRciS2hs2SAjXpqioxE4CUzaYkhb&confirm=t",
            self.dir,
            filename=f"PSM.zip",
        )

    def _load(self) -> np.ndarray:

        self.train_filepath = os.path.join(os.path.join(self.dir, "PSM"), 'train.csv')
        self.test_filepath = os.path.join(os.path.join(self.dir, "PSM"), 'test.csv')
        self.labels_filepath = os.path.join(os.path.join(self.dir, "PSM"), 'test_label.csv')
        
        self.train_data = np.nan_to_num(pd.read_csv(self.train_filepath).values[:, 1:])
        self.test_data = np.nan_to_num(pd.read_csv(self.test_filepath).values[:, 1:])
        self.test_labels = pd.read_csv(os.path.join(self.labels_filepath)).values[:, 1:]
        
        self.num_features = self.train_data.shape[1]
        




import os
import resource
from.dataset import Dataset, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import download_and_extract_archive, check_integrity, download_url
import pandas as pd
import numpy as np
import torch.utils.data



class Electricity(TimeSeriesDataset):
    """
    The raw dataset is available at https://archive.ics.uci.edu/ml/datasets/ElectricityLoadDiagrams20112014.
    It contains the electricity consumption in kWh recorded every 15 minutes from 2011 to 2014.
    Some dimensions were equal to 0, so we eliminated the records from 2011.
    The final dataset contains electricity consumption data for 321 clients from 2012 to 2014.
    The data was converted to reflect hourly consumption.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        freq (str): Frequency of the data points, in minutes.
        length (int): Length of the dataset.

    Methods:
        download():
            Downloads and extracts the dataset.
        _load():
            Loads the dataset into a NumPy array.
    """

    name:str= 'electricity'
    num_features: int = 321
    freq: str = 't' # in minuts
    length :int = 26304
    
    def download(self):
        download_url(
            "https://raw.githubusercontent.com/wayne155/multivariate_timeseries_datasets/main/electricity/electricity.csv",
            self.dir,
            filename="electricity.csv",
            md5="a1973ba4f4bed84136013ffa1ca27dc8",
        )
        
    def _load(self) -> np.ndarray:
        self.file_name = os.path.join(self.dir, 'electricity.csv')
        self.df = pd.read_csv(self.file_name, parse_dates=['date'])
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.drop("date", axis=1).values
        return self.data
    


import torch
from torch import Tensor
from typing import Generic, TypeVar, Union
from .base import Scaler, StoreType
import pandas as pd
import numpy as np
import torch

class MaxAbsScaler(Scaler[StoreType]):
    """
    shape of data :  (N , n)
    - N : sample num
    - n : node num
    Transforms each channel to the range [0, 1].
    """

    def __init__(self) -> None:
        self.scale = None

    def fit(self, data: StoreType):
        if isinstance(data, np.ndarray):
            self.scale = np.max(np.abs(data), axis=0)
        elif isinstance(data, Tensor):
            self.scale = data.abs().max(axis=0).values
        else:
            raise ValueError(f"not supported type : {type(data)}")

    def transform(self, data) -> StoreType:
        # (b , n)  or (n)
        return data / self.scale

    def inverse_transform(self, data: StoreType) -> StoreType:
        if isinstance(data, np.ndarray):
            return data * self.scale
        elif isinstance(data, Tensor):
            return data * torch.tensor(self.scale, device=data.device)
        else:
            raise ValueError(f"not supported type : {type(data)}")



import torch
from torch import Tensor
from typing import Generic, TypeVar, Union

import pandas as pd
import numpy as np
import torch
StoreType = TypeVar('StoreType')

class Scaler(Generic[StoreType]):
    """
    Generic Scaler class for fitting, transforming, and inverse transforming datasets.

    Methods:
        fit(data: StoreType) -> None:
            Fit the Scaler to the input dataset.

        transform(data: StoreType) -> StoreType:
            Transform the input dataset using the Scaler.

        inverse_transform(data: StoreType) -> StoreType:
            Perform an inverse transform on the input dataset using the Scaler.
    """

    def fit(self, data: StoreType) -> None:
        """
        Fit the Scaler to the input dataset.

        Args:
            data (StoreType): The input dataset to fit the Scaler to.

        Returns:
            None
        """
        raise NotImplementedError()

    def transform(self, data: StoreType) -> StoreType:
        """
        Transform the input dataset using the Scaler.

        Args:
            data (StoreType): The input dataset to transform using the Scaler.

        Returns:
            StoreType: The transformed dataset of the same type as the input data.
        """
        raise NotImplementedError()

    def inverse_transform(self, data: StoreType) -> StoreType:
        """
        Perform an inverse transform on the input dataset using the Scaler.

        Args:
            data (StoreType): The input dataset to perform an inverse transform on.

        Returns:
            StoreType: The inverse transformed dataset of the same type as the input data.
        """
        raise NotImplementedError()




import os
import resource
from.dataset import Dataset, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import download_url, download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np
import torch.utils.data


class Weather(TimeSeriesDataset):
    """
    This dataset contains local climatological data for nearly 1,600 in US locations, covering 4 years from 2010 to 2013. Data points are collected every hour. Each data point consists of the target value “wet bulb” and 11 climate features.
    The raw data can be acquired at https://www.ncei.noaa.gov/data/local-climatological-data/.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        length (int): Length of the dataset.
        freq (str): Frequency of the data points.

    Methods:
        download():
            Downloads and extracts the dataset.
        _load():
            Loads the dataset into a NumPy array.
    """
    
    name:str= 'weather'
    num_features: int = 21
    length : int  = 52696
    freq:str=  'h'
    
    
    def download(self):
        download_and_extract_archive(
         "https://raw.githubusercontent.com/wayne155/multivariate_timeseries_datasets/main/weather.zip",
         self.dir,
         filename='weather.zip',
         md5="fe40dc5c7e1787ad8ae63e2cadb5abfe"
        )
        
        
        
    def _load(self) -> np.ndarray:
        self.file_path =  os.path.join( os.path.join(self.dir, 'weather')  , 'weather.csv') 
        self.df = pd.read_csv(self.file_path,parse_dates=[0])
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.iloc[:, 1:].to_numpy()
        return self.data
    


from typing import Sequence, Tuple, Type

import numpy as np
import torch

from torch_timeseries.dataset import UEA
from torch_timeseries.utils.timefeatures import time_features
from ..scaler import Scaler
from ..dataset import (
    TimeSeriesDataset,
    TimeseriesSubset,
)
from torch.utils.data import Dataset, DataLoader, RandomSampler, Subset

from .wrapper import MultiStepTimeFeatureSet





class UEADataset(Dataset):
    def __init__(self, dataset: UEA, scaler: Scaler, flag: str ='train', window: int = 168, scaler_fit=True):
        self.dataset = dataset
        self.window = window
        self.scaler = scaler
        self.flag = flag
        if scaler_fit:
            self.scaler.fit(self.dataset.train_features_data)
        if self.flag == 'test':
            self.scaled_feature_df = self.scaler.transform(self.dataset.test_df)
            self.feature_df = self.dataset.test_df
            self.labels = self.dataset.test_labels
        elif self.flag == 'train':
            self.scaled_feature_df = self.scaler.transform(self.dataset.train_df)
            self.feature_df = self.dataset.train_df
            self.labels = self.dataset.train_labels
        
        self.indexes = self.feature_df.index.unique()
        
    def __getitem__(self, ind):
        scaled_x = torch.tensor(self.scaled_feature_df.loc[ind].values)
        x = torch.tensor(self.feature_df.loc[ind].values)
        y = torch.tensor(self.labels.loc[ind].values)
        return scaled_x, x, y
      
    def __len__(self):
        return len(self.indexes)




class UEAClassification:
    """
    Class for handling the classification of UEA datasets.

    Attributes:
        batch_size (int): Number of samples per batch.
        num_worker (int): Number of worker threads for data loading.
        dataset (UEA): UEA dataset to be used.
        scaler (Scaler): Scaler to normalize the data.
        window (int): Window size for the time series data. If not enough data, zeros will be used for padding.
        shuffle_train (bool): Whether to shuffle the training data.
    """
    def __init__(
        self,
        dataset: UEA,
        scaler: Scaler,
        window: int = 168,
        shuffle_train=True,
        batch_size: int = 32,
        num_worker: int = 3,
    ) -> None:

        self.batch_size = batch_size
        self.num_worker = num_worker
        self.dataset = dataset

        self.scaler = scaler
        self.window = window
        self.shuffle_train = shuffle_train

        self._load()

    def _load(self):
        self._load_dataset()
        self._load_dataloader()

    def _load_dataset(self):
        """
        Return the splitted training, testing and validation dataloders

        :return: a tuple of train_dataloader, test_dataloader and val_dataloader
        """
        # fixed suquence dataset
        self.train_dataset = UEADataset(self.dataset, self.scaler, 'train', self.window, scaler_fit=True)
        self.val_dataset = UEADataset(self.dataset, self.scaler, 'test', self.window, scaler_fit=False)
        self.test_dataset = UEADataset(self.dataset, self.scaler, 'test', self.window, scaler_fit=False)

    def _load_dataloader(self):
        self.train_size = len(self.train_dataset)
        self.val_size = len(self.val_dataset)
        self.test_size = len(self.test_dataset)
        self.train_loader = DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_worker,
            drop_last=False,
            collate_fn=lambda x: collate_fn(x, max_len=self.window)
        )

        self.val_loader = DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_worker,
            drop_last=False,
            collate_fn=lambda x: collate_fn(x, max_len=self.window)
        )

        self.test_loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_worker,
            drop_last=False,
            collate_fn=lambda x: collate_fn(x, max_len=self.window)
        )







def collate_fn(data, max_len=None):
    """Build mini-batch tensors from a list of (X, mask) tuples. Mask input. Create
    Args:
        data: len(batch_size) list of tuples (X, y).
            - X: torch tensor of shape (seq_length, feat_dim); variable seq_length.
            - y: torch tensor of shape (num_labels,) : class indices or numerical targets
                (for classification or regression, respectively). num_labels > 1 for multi-task models
        max_len: global fixed sequence length. Used for architectures requiring fixed length input,
            where the batch length cannot vary dynamically. Longer sequences are clipped, shorter are padded with 0s
    Returns:
        X: (batch_size, padded_length, feat_dim) torch tensor of masked features (input)
        targets: (batch_size, padded_length, feat_dim) torch tensor of unmasked features (output)
        target_masks: (batch_size, padded_length, feat_dim) boolean torch tensor
            0 indicates masked values to be predicted, 1 indicates unaffected/"active" feature values
        padding_masks: (batch_size, padded_length) boolean tensor, 1 means keep vector at this position, 0 means padding
    """
    batch_size = len(data)
    features, raw_features, labels = zip(*data)

    # Stack and pad features and masks (convert 2D to 3D tensors, i.e. add batch dimension)
    lengths = [X.shape[0] for X in features]  # original sequence length for each time series
    if max_len is None:
        max_len = max(lengths)

    X = torch.zeros(batch_size, max_len, features[0].shape[-1])  # (batch_size, padded_length, feat_dim)
    scaled_X = torch.zeros(batch_size, max_len, features[0].shape[-1])  # (batch_size, padded_length, feat_dim)
    for i in range(batch_size):
        end = min(lengths[i], max_len)
        scaled_X[i, :end, :] = features[i][:end, :]
        X[i, :end, :] = raw_features[i][:end, :]

    targets = torch.stack(labels, dim=0)  # (batch_size, num_labels)

    padding_masks = padding_mask(torch.tensor(lengths, dtype=torch.int16),
                                 max_len=max_len)  # (batch_size, padded_length) boolean tensor, "1" means keep

    return scaled_X, X, targets, padding_masks



def padding_mask(lengths, max_len=None):
    """
    Used to mask padded positions: creates a (batch_size, max_len) boolean mask from a tensor of sequence lengths,
    where 1 means keep element at this position (time step)
    """
    batch_size = lengths.numel()
    max_len = max_len or lengths.max_val()  # trick works because of overloading of 'or' operator for non-boolean types
    return (torch.arange(0, max_len, device=lengths.device)
            .type_as(lengths)
            .repeat(batch_size, 1)
            .lt(lengths.unsqueeze(1)))





import numpy as np
import pandas as pd
from torch_timeseries.dataset.dataset import Freq, TimeSeriesDataset, TimeSeriesStaticGraphDataset


class Dummy(TimeSeriesDataset):
    """
    Dummy dataset for testing purposes.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        freq (Freq): Frequency of the data points.
        length (int): Length of the dataset.

    Methods:
        download():
            Placeholder method for downloading data.
        _load():
            Loads the dataset into a NumPy array.
    """

    name: str = 'dummy'
    num_features:int = 2
    freq : Freq = Freq.minutes
    length : int = 1440
    def download(self): 
        pass
    
    def _load(self):
        dates = pd.date_range(start='2022-01-01', end='2022-01-02', freq='t')

        data = np.random.rand(len(dates), 2)
        self.df = pd.DataFrame({'date': dates, 'data1': data[:, 0],'data2': data[:, 1]})
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.drop('date', axis=1).values        
        return self.data
    
    
    


import os
import resource
from .dataset import Dataset, TimeSeriesDataset
from typing import Callable, List, Optional
import torch
from torchvision.datasets.utils import download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np

class Traffic(TimeSeriesDataset):
    """Traffic dataset for road occupancy rates in the San Francisco Bay area.

    The raw data is sourced from the California Department of Transportation
    and is available at http://pems.dot.ca.gov. This dataset contains 48 months
    (2015-2016) of hourly road occupancy rates (between 0 and 1) measured by
    different sensors on freeways in the San Francisco Bay area.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        length (int): Length of the dataset.
        freq (str): Frequency of the data points.

    Methods:
        download(): Downloads and extracts the dataset.
        _load(): Loads the dataset into a NumPy array.
    """

    name:str= 'traffic'
    num_features: int = 862
    length : int = 17544
    freq:str = 'h'
    
    def download(self) -> None:
        download_and_extract_archive(
            "https://raw.githubusercontent.com/laiguokun/multivariate-time-series-data/master/traffic/traffic.txt.gz",
            self.dir,
            filename="traffic.txt.gz",
            md5="db745d0c9f074159581a076cbb3f23d6",
        )
        
    def _load(self) -> np.ndarray:
        self.file_name = os.path.join(self.dir, 'traffic.txt')
        self.df = pd.read_csv(self.file_name, header=None)
        self.df['date'] = pd.date_range(start='01/01/2015 00:00', periods=self.length, freq='1H')  # '5T' for 5 minutes
        self.dates =  pd.DataFrame({'date': self.df['date'] })

        # self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.drop("date", axis=1).values
        # self.data = np.loadtxt(self.file_name, delimiter=',')
        return self.data
        


import numpy as np
import pandas as pd
from typing import Any, Callable, Generic, NewType, Optional, Sequence, TypeVar, Union
import torch.utils.data
import os
from abc import ABC, abstractmethod


from enum import Enum, unique


@unique
class Freq(str, Enum):
    seconds = "s"
    minutes = "t"
    hours = "h"
    days = "d"
    months = "m"
    years = "y"


class Dataset(torch.utils.data.Dataset):
    name: str
    num_features: int
    length: int
    freq: Freq

    def __init__(self, root: str):
        """_summary_

        Args:
            root (str): data save location
            transform (Optional[Callable], optional): _description_. Defaults to None.
            target_transform (Optional[Callable], optional): _description_. Defaults to None.
            single_step (bool, optional): True for single_step data, False for multi_steps data. Defaults to True.
        """
        super().__init__()

        self.data: np.ndarray
        self.df: pd.DataFrame

    def download(self):
        r"""Downloads the dataset to the :obj:`self.dir` folder."""
        raise NotImplementedError

    def __len__(self):
        return len(self.data)


# StoreTypes = Union[np.ndarray, Tensor]
StoreTypes = np.ndarray


class TimeSeriesDataset(Dataset):
    def __init__(self, root: str='./data'):
        """_summary_

        Args:
            root (str): data save location

        """
        super().__init__(root)
        self.root = root
        self.dir = os.path.join(root, self.name)
        os.makedirs(self.dir, exist_ok=True)
        
        self.download()
        self._process()
        self._load()
        
        self.dates: Optional[pd.DataFrame]
        

    @abstractmethod
    def download(self):
        r"""Downloads the dataset to the :obj:`self.dir` folder."""
        raise NotImplementedError

    def _process(self) :
        pass
    
    
    @abstractmethod
    def _load(self) -> StoreTypes:
        """Loads the dataset to the :attr:`self.data` .

        Raises:
            NotImplementedError: _description_

        Returns:
            T: should return a numpy.array or torch.tensor or pandas.Dataframe
        """

        raise NotImplementedError



class AnomalyDataset(TimeSeriesDataset):
    train_data : Optional[np.array]
    test_data : Optional[np.array]
    test_labels : Optional[np.array]
        
        

class TimeSeriesStaticGraphDataset(TimeSeriesDataset):
    adj : np.ndarray 
    def _load_static_graph(self):
        raise NotImplementedError()


class TimeseriesSubset(torch.utils.data.Subset):
    def __init__(self, dataset: TimeSeriesDataset, indices: Sequence[int]) -> None:
        self.dataset = dataset
        self.indices = indices
        self.data = self.dataset.data[indices]
        self.df = self.dataset.df.iloc[indices]
        self.dates = self.dataset.dates.iloc[indices]
        self.num_features = dataset.num_features
        self.name = dataset.name
        self.length = len(self.indices)
        self.freq = dataset.freq





import numpy as np
import pandas as pd
from torch_timeseries.dataset.dataset import Freq, TimeSeriesDataset, TimeSeriesStaticGraphDataset


class DummyGraph(TimeSeriesStaticGraphDataset):
    """
    Dummy graph dataset for testing purposes.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        freq (Freq): Frequency of the data points.
        length (int): Length of the dataset.

    Methods:
        _load_static_graph():
            Loads a static adjacency matrix for the graph.
        download():
            Placeholder method for downloading data.
        _load():
            Loads the dataset into a NumPy array.
    """


    name: str = 'dummy_graph'
    num_features:int = 5
    freq : Freq = Freq.minutes
    length : int = 1440
    
    def _load_static_graph(self):
        self.adj = np.ones((self.num_features, self.num_features))
    def download(self): 
        pass
    
    def _load(self):
        self._load_static_graph()
        dates = pd.date_range(start='2022-01-01',periods=self.length, freq='t')

        data = np.random.rand(len(dates), self.num_features)
        result = {'date': dates}
        # iterate to get above df
        for i in range(data.shape[1]):  
            key = f'data{i+1}' 
            result[key] = data[:, i] 
        self.df = pd.DataFrame(result)

        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.drop('date', axis=1).values        
        return self.data




import pandas as pd
from torchvision.datasets.utils import download_and_extract_archive, check_integrity
import torch
from typing import Callable, List, Optional
import os
import resource
import numpy as np
from .dataset import Dataset, TimeSeriesDataset


class SolarEnergy(TimeSeriesDataset):
    """
    The dataset contains solar power production records for the year 2006, sampled every 5 minutes from 137 PV plants in Alabama State.
    The raw data is available at http://www.nrel.gov/grid/solar-power-data.html.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        length (int): Length of the dataset.
        freq (str): Frequency of the data points.
        file_name (str): Name of the file containing the dataset.

    Methods:
        download():
            Downloads and extracts the dataset.
        _load():
            Loads the dataset into a NumPy array.
    """

    name: str = "solar_AL"
    num_features: int = 137
    length: int = 52560
    freq:str = 'h'
    
    file_name = "solar_AL.txt"

    def download(self) -> None:
        # download_and_extract_archive(
        #     "https://www.nrel.gov/grid/assets/downloads/al-pv-2006.zip",
        #     self.dir,
        #     filename="al-pv-2006.zip",
        #     md5="3fa6015aa550fc1f50d2f9bd6909403c",
        # )
        download_and_extract_archive(
            "https://raw.githubusercontent.com/laiguokun/multivariate-time-series-data/master/solar-energy/solar_AL.txt.gz",
            self.dir,
            filename="solar_AL.txt.gz",
            md5="41ef7fdc958c2ca3fac9cd06d6227073",
        )

    # def _check_procssed(self):
    #     self.file_path = os.path.join(self.dir, self.file_name)
    #     return check_integrity(self.file_path)

    # def _process(self) -> None:
    #     if not self._check_procssed():
    #         print(f"processing {self.name} files")
    #         all_files = os.listdir(self.dir)
    #         csv_files = [
    #             filename
    #             for filename in all_files
    #             if filename.endswith(".csv") and filename.endswith("5_Min.csv")
    #         ]
    #         df_list = []
    #         with tqdm(total=len(csv_files)) as progress_bar:
    #             for i, csv_file in enumerate(csv_files):
    #                 file_path = os.path.join(self.dir, csv_file)
    #                 df = pd.read_csv(file_path, index_col=0, parse_dates=True)
    #                 column_name = f"Power{len(df_list)+1}(MW)"
    #                 df = df.rename(columns={"Power(MW)": column_name})
    #                 df_list.append(df[column_name])

    #                 progress_bar.update(1)
    #             merged_df = pd.concat(df_list, axis=1)
    #             merged_df.to_csv(os.path.join(self.dir, self.file_name))
    #     else:
    #         print("Using processed data ...")

    def _load(self) -> np.ndarray:
        self.file_path = os.path.join(self.dir, self.file_name)
        self.df = pd.read_csv(self.file_path, sep=',', header=None) # pd.read_csv(self.file_path, parse_dates=["LocalTime"])
        # self.df = df.rename(columns={"LocalTime": "date"})
        self.df['date'] = pd.date_range(start='01/01/2006 00:00', periods=self.length, freq='10T')  #
        self.dates =  pd.DataFrame({'date': self.df['date'] })
        self.data = self.df.drop("date", axis=1).values
        return self.data




import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_timeseries.nn.kernels import MovingAvg

class SeriesDecompMulti(nn.Module):
    """
    Series decomposition block
    """
    def __init__(self, kernel_size):
        super(SeriesDecompMulti, self).__init__()
        self.moving_avg = [MovingAvg(kernel, stride=1) for kernel in kernel_size]
        self.layer = torch.nn.Linear(1, len(kernel_size))

    def forward(self, x):
        moving_mean=[]
        for func in self.moving_avg:
            moving_avg = func(x)
            moving_mean.append(moving_avg.unsqueeze(-1))
        moving_mean=torch.cat(moving_mean,dim=-1)
        moving_mean = torch.sum(moving_mean*nn.Softmax(-1)(self.layer(x.unsqueeze(-1))),dim=-1)
        res = x - moving_mean
        return res, moving_mean 

class SeriesDecomp(nn.Module):
    """
    Series decomposition block
    """
    def __init__(self, kernel_size):
        super(SeriesDecomp, self).__init__()
        self.moving_avg = MovingAvg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


import torch
from torch import Tensor
from typing import Generic, TypeVar, Union
from .base import Scaler, StoreType
import pandas as pd
import numpy as np
import torch


class StandardScaler(Scaler):
    """
    shape of data :  (N , n)
    - N : sample num
    - n : node num
    Transforms each channel to the range [0, 1].
    """

    def __init__(self, device="cpu") -> None:
        self.mean = None
        self.std = None
        self.device = device

    def fit(self, data: StoreType):
        if isinstance(data, np.ndarray):
            self.mean = np.mean(data, axis=0)
            self.std = np.std(data, axis=0)
            # do not normalize all zero values
            self.std[self.std == 0] = 1
        elif isinstance(data, Tensor):
            self.mean = data.mean(axis=0)
            self.std = data.std(axis=0)
            self.std[self.std == 0] = 1
        else:
            raise ValueError(f"not supported type : {type(data)}")

    def transform(self, data):
        return (data - self.mean) / self.std

    def inverse_transform(self, data: StoreType) -> StoreType:
        if isinstance(data, np.ndarray):
            return data * self.std + self.mean
        elif isinstance(data, Tensor):
            return data * torch.tensor(self.std, device=data.device) + torch.tensor(
                self.mean, device=data.device
            )
        else:
            raise ValueError(f"not supported type : {type(data)}")



import os
import resource
from .dataset import Dataset, Freq, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import (
    download_url,
    download_and_extract_archive,
    check_integrity,
)
import pandas as pd
import numpy as np
import torch.utils.data


class ETTh2(TimeSeriesDataset):
    """
    Electricity Transformer Temperature hourly. This dataset contains 2-year data from two separate counties in China. Each data point consists of the target value "oil temperature" and 6 power load features.
    
    This dataset is collected by Informer and is available at https://github.com/zhouhaoyi/ETDataset.

    Attributes:
        name (str): Name of the dataset.
        freq (str): Frequency of the data points.
        num_features (int): Number of features in the dataset.
        length (int): Length of the dataset.

    Methods:
        download():
            Downloads and extracts the dataset.
        _load():
            Loads the dataset into a NumPy array.
        _process():
            Processes the dataset using the parent class's process method.
    """

    name: str = "ETTh2"
    freq: Freq = "h"
    num_features: int = 7
    length: int = 17420

    def download(self):
        download_url(
            "https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTh2.csv",
            self.dir,
            filename="ETTh2.csv",
            md5="51a229a3fc13579dd939364fefe9c7ab",
        )

    def _load(self) -> np.ndarray:
        self.file_path = os.path.join(self.dir, "ETTh2.csv")
        self.df = pd.read_csv(self.file_path, parse_dates=[0])
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.iloc[:, 1:].to_numpy()
        return self.data

    def _process(self) -> np.ndarray:
        return super()._process()



# https://github.com/Mcompetitions/M4-methods/tree/master/Dataset


import os
import resource
from .dataset import Dataset, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import (
    download_url,
    download_and_extract_archive,
    check_integrity,
)
import pandas as pd
import numpy as np
import torch.utils.data


class M4(TimeSeriesDataset):
    """
    The M4 forecasting competition, the continuation of the previous three ones organized by Spyros Makridakis (https://en.wikipedia.org/wiki/Makridakis_Competitions).
    collected from https://github.com/Mcompetitions/M4-methods/tree/master,we concate train,test files as the dataset.

    Parameters:
    category (str): The category of the dataset to load. Can be one of ['Daily', 'Hourly', 'Monthly', 'Quarterly', 'Weekly', 'Yearly']. Default is 'Daily'.
    root (str): The root directory to store the dataset. Default is './data'.

    Attributes:
    dir (str): The directory where the dataset is stored.
    dates (pd.DataFrame): The dates associated with the dataset.
    df (pd.DataFrame): The combined training and test data.
    data (np.ndarray): The time series data.

    Methods:
    download(): Download the dataset from the M4 repository.
    _load(): Load the dataset from the local files.
    """


    name: str = "M4"

    def __init__(self, root: str = "./data", category="Daily"):
        self.category = category
        self.root = root
        self.dir = os.path.join(root, self.name)
        os.makedirs(self.dir, exist_ok=True)

        self.download()
        self._process()
        self._load()

        self.dates: Optional[pd.DataFrame]

    def _download_category(self, category):
        download_url(
            f"https://raw.githubusercontent.com/Mcompetitions/M4-methods/master/Dataset/Train/{category}-train.csv",
            self.dir,
            filename=f"{category}-train.csv",
        )

        download_url(
            f"https://raw.githubusercontent.com/Mcompetitions/M4-methods/master/Dataset/Test/{category}-test.csv",
            self.dir,
            filename=f"{category}-test.csv",
        )
    def download(self):
        self._download_category(self.category)

    def _load(self) -> np.ndarray:

        self.train_file_path = os.path.join(self.dir, f"{self.category}-train.csv")
        self.test_file_path = os.path.join(self.dir, f"{self.category}-test.csv")

        self.tarin_df = pd.read_csv(self.train_file_path)
        self.test_df = pd.read_csv(self.test_file_path)
        self.df = pd.concat([self.tarin_df, self.test_df]).dropna(axis=1)
        self.data = self.df.iloc[:, 1:].to_numpy()
        return self.data



import os
import resource
from .dataset import Dataset, TimeSeriesDataset, AnomalyDataset
from typing import Callable, List, Optional
import torch
from torchvision.datasets.utils import download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np

class MSL(AnomalyDataset):
    """
    MSL is a public dataset from NASA with 55 dimensions showing the status of sensor and actuator data from the Mars Rover.
    """

    name:str= 'MSL'
    
    def download(self) -> None:
        download_and_extract_archive(
            f"https://drive.usercontent.google.com/download?id=14STjpszyi6D0B7BUHZ1L4GLUkhhPXE0G&confirm=t",
            self.dir,
            filename=f"MSL.zip",
        )

    def _load(self) -> np.ndarray:
        self.train_filepath = os.path.join(os.path.join(self.dir, "MSL"), 'MSL_train.npy')
        self.test_filepath = os.path.join(os.path.join(self.dir, "MSL"), 'MSL_test.npy')
        self.labels_filepath = os.path.join(os.path.join(self.dir, "MSL"), 'MSL_test_label.npy')
        
        self.train_data = np.load(self.train_filepath)
        self.test_data = np.load(self.test_filepath)
        self.test_labels = np.load(self.labels_filepath)
        
        self.num_features = self.train_data.shape[1]
        
        




import os
import resource
from .dataset import Dataset, TimeSeriesDataset
from typing import Callable, List, Optional
import torch
from torchvision.datasets.utils import download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np

class SWaT(TimeSeriesDataset):
    """
    SWaT (Safe Water Treatment, Mathur & Tippenhauer (2016)) is obtained from 51 sensors in continuously operating critical infrastructure systems.
    """

    name:str= 'SWaT'
    
    def download(self) -> None:
        download_and_extract_archive(
            f"https://drive.usercontent.google.com/download?id=1eRKQwJhqmUD4LkWnqNy1cdIz3W_y6EtW&confirm=t",
            self.dir,
            filename=f"SWaT.zip",
        )

    def _load(self) -> np.ndarray:
        self.train_filepath = os.path.join(os.path.join(self.dir, "SWaT"), 'swat_train2.csv')
        self.test_filepath = os.path.join(os.path.join(self.dir, "SWaT"), 'swat2.csv')
        
        self.train_df = pd.read_csv(self.train_filepath)
        self.test_df = pd.read_csv(self.test_filepath)
        
        self.train_data, self.train_labels = self.train_df.values[:, :-1], self.train_df.values[:, -1:]
        self.test_data, self.test_labels = self.test_df.values[:, :-1], self.test_df.values[:, -1:]
        
        self.num_features = self.train_data.shape[1]
        


from typing import Sequence, Tuple, Type

import torch
from ..scaler import Scaler
from torch_timeseries.dataset import (
    TimeSeriesDataset,
    TimeseriesSubset,
)
from torch.utils.data import Dataset, DataLoader, RandomSampler, Subset

from .wrapper import MultiStepTimeFeatureSet, MultiStepSet


class SlidingWindow:
    def __init__(
        self,
        dataset: TimeSeriesDataset,
        scaler: Scaler,
        window: int = 168,
        horizon: int = 3,
        steps: int = 2,
        scale_in_train=False,
        shuffle_train=True,
        batch_size: int = 32,
        train_ratio: float = 0.7,
        val_ratio: float = 0.2,
        num_worker: int = 3,
        uniform_eval=True,
    ) -> None:
        """
        Class for splitting the dataset sequentially and then randomly sampling from each subset.

        Attributes:
            dataset (TimeSeriesDataset): Time series dataset to be used.
            scaler (Scaler): Scaler to normalize the data.
            window (int): Window size for the time series data.
            horizon (int): Forecast horizon.
            steps (int): Step size between windows.
            scale_in_train (bool): Whether to scale data during training.
            shuffle_train (bool): Whether to shuffle the training data.
            batch_size (int): Number of samples per batch.
            train_ratio (float): Ratio of the dataset to be used for training.
            val_ratio (float): Ratio of the dataset to be used for validation.
            test_ratio (float): Ratio of the dataset to be used for testing.
            num_worker (int): Number of worker threads for data loading.
            uniform_eval (bool): Whether to use uniform evaluation.
            train_loader (DataLoader): DataLoader for the training data.
            val_loader (DataLoader): DataLoader for the validation data.
            test_loader (DataLoader): DataLoader for the test data.
        """
        self.train_ratio = train_ratio
        self.val_ratio = val_ratio
        self.test_ratio = 1 - self.train_ratio - self.val_ratio
        self.uniform_eval = uniform_eval

        assert (
            self.train_ratio + self.val_ratio + self.test_ratio == 1.0
        ), "Split ratio must sum up to 1.0"
        self.batch_size = batch_size
        self.num_worker = num_worker
        self.dataset = dataset

        self.scaler = scaler
        self.window = window
        self.steps = steps
        self.horizon = horizon
        self.shuffle_train = shuffle_train
        self.scale_in_train = scale_in_train

        self._load()

    def _load(self):
        self._load_dataset()
        self._load_dataloader()

    def _load_dataset(self):
        """
        Return the splitted training, testing and validation dataloders

        :return: a tuple of train_dataloader, test_dataloader and val_dataloader
        """
        # fixed suquence dataset
        indices = range(0, len(self.dataset))

        train_size = int(self.train_ratio * len(self.dataset))
        val_size = int(self.val_ratio * len(self.dataset))
        test_size = len(self.dataset) - val_size - train_size
        train_subset = TimeseriesSubset(self.dataset, indices[0:train_size])
        
        if self.uniform_eval:
            val_subset = TimeseriesSubset( # self.window + self.horizon - 1
                self.dataset, indices[train_size - self.window - self.horizon + 1: (val_size + train_size)]
            )
        else:
            val_subset = TimeseriesSubset(
                self.dataset,indices[train_size: (val_size + train_size)]
            )

        if self.uniform_eval:
            test_subset = TimeseriesSubset(self.dataset, indices[-test_size - self.window - self.horizon + 1:])
        else:
            test_subset = TimeseriesSubset(self.dataset, indices[-test_size:])

            
        if self.scale_in_train:
            self.scaler.fit(train_subset)
        else:
            self.scaler.fit(self.dataset.data)

        self.train_dataset = MultiStepSet(
            train_subset,
            scaler=self.scaler,
            window=self.window,
            horizon=self.horizon,
            steps=self.steps,
            scaler_fit=False,
        )
        self.val_dataset = MultiStepSet(
            val_subset,
            scaler=self.scaler,
            window=self.window,
            horizon=self.horizon,
            steps=self.steps,
            scaler_fit=False,
        )
        self.test_dataset = MultiStepSet(
            test_subset,
            scaler=self.scaler,
            window=self.window,
            horizon=self.horizon,
            steps=self.steps,
            scaler_fit=False,
        )

    def _load_dataloader(self):
        self.train_size = len(self.train_dataset)
        self.val_size = len(self.val_dataset)
        self.test_size = len(self.test_dataset)
        self.train_loader = DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=self.shuffle_train,
            num_workers=self.num_worker,
        )

        self.val_loader = DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_worker,
        )

        self.test_loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_worker,
        )




from typing import List

import numpy as np
import pandas as pd
from pandas.tseries import offsets
from pandas.tseries.frequencies import to_offset
pd.options.mode.chained_assignment = None
class TimeFeature:
    def __init__(self):
        pass

    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        pass

    def __repr__(self):
        return self.__class__.__name__ + "()"

class SecondOfMinute(TimeFeature):
    """Minute of hour encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return index.second / 59.0 - 0.5

class MinuteOfHour(TimeFeature):
    """Minute of hour encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return index.minute / 59.0 - 0.5

class HourOfDay(TimeFeature):
    """Hour of day encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return index.hour / 23.0 - 0.5

class DayOfWeek(TimeFeature):
    """Hour of day encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return index.dayofweek / 6.0 - 0.5

class DayOfMonth(TimeFeature):
    """Day of month encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return (index.day - 1) / 30.0 - 0.5

class DayOfYear(TimeFeature):
    """Day of year encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return (index.dayofyear - 1) / 365.0 - 0.5

class MonthOfYear(TimeFeature):
    """Month of year encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return (index.month - 1) / 11.0 - 0.5

class WeekOfYear(TimeFeature):
    """Week of year encoded as value between [-0.5, 0.5]"""
    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        return (index.week - 1) / 52.0 - 0.5

def time_features_from_frequency_str(freq_str: str) -> List[TimeFeature]:
    """
    Returns a list of time features that will be appropriate for the given frequency string.
    Parameters
    ----------
    freq_str
        Frequency string of the form [multiple][granularity] such as "12H", "5min", "1D" etc.
    """

    features_by_offsets = {
        offsets.YearEnd: [],
        offsets.QuarterEnd: [MonthOfYear],
        offsets.MonthEnd: [MonthOfYear],
        offsets.Week: [DayOfMonth, WeekOfYear],
        offsets.Day: [DayOfWeek, DayOfMonth, DayOfYear],
        offsets.BusinessDay: [DayOfWeek, DayOfMonth, DayOfYear],
        offsets.Hour: [HourOfDay, DayOfWeek, DayOfMonth, DayOfYear],
        offsets.Minute: [
            MinuteOfHour,
            HourOfDay,
            DayOfWeek,
            DayOfMonth,
            DayOfYear,
        ],
        offsets.Second: [
            SecondOfMinute,
            MinuteOfHour,
            HourOfDay,
            DayOfWeek,
            DayOfMonth,
            DayOfYear,
        ],
    }

    offset = to_offset(freq_str)

    for offset_type, feature_classes in features_by_offsets.items():
        if isinstance(offset, offset_type):
            return [cls() for cls in feature_classes]

    supported_freq_msg = f"""
    Unsupported frequency {freq_str}
    The following frequencies are supported:
        Y   - yearly
            alias: A
        M   - monthly
        W   - weekly
        D   - daily
        B   - business days
        H   - hourly
        T   - minutely
            alias: min
        S   - secondly
    """
    raise RuntimeError(supported_freq_msg)

def time_features(dates:pd.DataFrame, timeenc=1, freq='h') -> np.ndarray:
    """
    encode time features based on data sample freqency
    > `time_features` takes in a `dates` dataframe with a 'dates' column and extracts the date down to `freq` where freq can be any of the following if `timeenc` is 0: 
    > * m - [month]
    > * w - [month]
    > * d - [month, day, weekday]
    > * b - [month, day, weekday, ]
    > * h - [month, day, weekday, hour]
    > * t - [month, day, weekday, hour, *minute]
    > 
    > If `timeenc` is 1, a similar, but different list of `freq` values are supported (all encoded between [-0.5 and 0.5]): 
    > * Q - [month]
    > * M - [month]
    > * W - [Day of month, week of year]
    > * D - [Day of week, day of month, day of year]
    > * B - [Day of week, day of month, day of year]
    > * H - [Hour of day, day of week, day of month, day of year]
    > * T - [Minute of hour*, hour of day, day of week, day of month, day of year]
    > * S - [Second of minute, minute of hour, hour of day, day of week, day of month, day of year]

    *minute returns a number from 0-3 corresponding to the 15 minute period it falls into.
    """
    if timeenc==0:
        dates['month'] = dates.date.apply(lambda row:row.month,1)
        dates['day'] = dates.date.apply(lambda row:row.day,1)
        dates['weekday'] = dates.date.apply(lambda row:row.weekday(),1)
        dates['hour'] = dates.date.apply(lambda row:row.hour,1)
        dates['minute'] = dates.date.apply(lambda row:row.minute,1)
        dates['minute'] = dates.minute.map(lambda x:x//15)
        freq_map = {
            'y':[],'m':['month'],'w':['month'],'d':['month','day','weekday'],
            'b':['month','day','weekday'],'h':['month','day','weekday','hour'],
            't':['month','day','weekday','hour','minute'],
        }
        return dates[freq_map[freq.lower()]].values
    if timeenc==1:
        dates = pd.to_datetime(dates.date.values)
        return np.vstack([feat(dates) for feat in time_features_from_frequency_str(freq)]).transpose(1,0)



from setuptools import find_packages, setup

__version__ = '0.0.3'
URL = 'https://github.com/wayne155/pytorch_timeseries'

install_requires = [
    'tqdm',
    'torch',
    'numpy' ,
    'pandas',
    'torchvision',
    'tqdm',
    'sktime>=0.29.0',
    'scikit-learn',
]

full_requires =  [
    
]

benchmark_requires = [

]

test_requires = [
    'pytest',
]

dev_requires = test_requires + [
    'pre-commit',
]

setup(
    name='torch_timeseries',
    version=__version__,
    description='Timeseries Learning Library for PyTorch',
    author='Weiwei Ye',
    author_email='wwye155@gmail.com',
    url=URL,
    download_url=f'{URL}/archive/{__version__}.tar.gz',
    keywords=[
        'deep-learning',
        'pytorch',
        'timeseries-deep-learning',
        'neural-networks',
    ],
    python_requires='>=3.8',
    install_requires=install_requires,
    extras_require={
        'full': full_requires,
        'benchmark': benchmark_requires,
        'test': test_requires,
        'dev': dev_requires,
    },
    package_dir={"torch_timeseries":"torch_timeseries"},
    packages=find_packages(exclude=['test']),
    include_package_data=True,
)



import numpy as np
import pandas as pd
from torchvision.datasets.utils import download_and_extract_archive, check_integrity, download_url
import torch
from typing import Any, Callable, List, Optional
import os
import resource
from.dataset import Dataset, TimeSeriesDataset


class ExchangeRate(TimeSeriesDataset):
    """
    The collection of the daily exchange rates of eight foreign countries including Australia, British, Canada, Switzerland, China, Japan, New Zealand, and Singapore ranging from 1990 to 2016.
    
    The raw data is collected from https://github.com/laiguokun/multivariate-time-series-data.

    Attributes:
        name (str): Name of the dataset.
        num_features (int): Number of features in the dataset.
        freq (str): Frequency of the data points (daily).
        length (int): Length of the dataset.

    Methods:
        download():
            Downloads and extracts the dataset.
        _load():
            Loads the dataset into a NumPy array.
    """

    name: str = 'ExchangeRate'
    num_features: int = 8
    freq: str = 'd' # daily data
    length : int = 7588
    
    
    def download(self) -> None:
        download_url(
            "https://raw.githubusercontent.com/wayne155/multivariate_timeseries_datasets/main/exchange_rate/exchange_rate.csv",
            self.dir,
            filename="exchange_rate.csv",
            md5="2fc11972378a4c8817c1adfdde522bf9",
        )

    def _load(self) -> np.ndarray:
        self.file_name = os.path.join(self.dir, 'exchange_rate.csv')
        self.df = pd.read_csv(self.file_name, parse_dates=['date'])
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.drop("date", axis=1).values
        return self.data



from typing import Optional, TypeVar, Tuple

import pandas as pd
from ..scaler import Scaler

from torch_timeseries.utils.timefeatures import time_features
from ..dataset import TimeSeriesDataset, TimeseriesSubset
from torch.utils.data import Dataset
import numpy as np
import torch


class MultiStepTimeFeatureSet(Dataset):
    def __init__(self, dataset: TimeseriesSubset, scaler: Scaler, time_enc=0, window: int = 168, horizon: int = 3, steps: int = 2, freq=None, scaler_fit=True):
        self.dataset = dataset
        self.window = window
        self.horizon = horizon
        self.steps = steps
        self.time_enc = time_enc
        self.scaler = scaler
        
        self.num_features = self.dataset.num_features
        self.length = self.dataset.length
        if freq is None:
            self.freq = self.dataset.freq
        else:
            self.freq = freq
        if scaler_fit:
            self.scaler.fit(self.dataset.data)
        self.scaled_data = self.scaler.transform(self.dataset.data)
        self.date_enc_data = time_features(
            self.dataset.dates, self.time_enc, self.freq)
        assert len(self.dataset) - self.window - self.horizon + 1 - self.steps + 1 > 0, "Dataset is not long enough!!!"


    def transform(self, values):
        return self.scaler.transform(values)

    def inverse_transform(self, values):
        return self.scaler.inverse_transform(values)

    def __getitem__(self, index):
        # scaled_x : (B, T, N)
        # scaled_y : (B, O, N)
        # x : (B, T, N)
        # y : (B, O, N)
        # x_date_enc : (B, T, D)
        # y_date_eDc : (B, O, D)
        if isinstance(index, int):
            scaled_x = self.scaled_data[index:index+self.window]

            scaled_y = self.scaled_data[self.window + self.horizon - 1 +
                                 index:self.window + self.horizon - 1 + index+self.steps]
            
            x = self.dataset.data[index:index+self.window]
            
            y = self.dataset.data[self.window + self.horizon - 1 +
                                 index:self.window + self.horizon - 1 + index+self.steps]

            x_date_enc = self.date_enc_data[index:index+self.window]

            y_date_enc = self.date_enc_data[self.window + self.horizon -
                                            1 + index:self.window + self.horizon - 1 + index+self.steps]
            return scaled_x, scaled_y, x , y, x_date_enc, y_date_enc
        else:
            raise TypeError('Not surpported index type!!!')

    def __len__(self):
        return len(self.dataset) - self.window - self.horizon + 1 - self.steps + 1






class MultiStepSet(Dataset):
    def __init__(self, dataset: TimeseriesSubset, scaler: Scaler, window: int = 168, horizon: int = 3, steps: int = 2, scaler_fit=True):
        self.dataset = dataset
        self.window = window
        self.horizon = horizon
        self.steps = steps
        self.scaler = scaler
        
        self.num_features = self.dataset.num_features
        self.length = self.dataset.length
        if scaler_fit:
            self.scaler.fit(self.dataset.data)
        self.scaled_data = self.scaler.transform(self.dataset.data)
        assert len(self.dataset) - self.window - self.horizon + 1 - self.steps + 1 > 0, "Dataset is not long enough!!!"


    def transform(self, values):
        return self.scaler.transform(values)

    def inverse_transform(self, values):
        return self.scaler.inverse_transform(values)

    def __getitem__(self, index):
        # scaled_x : (B, T, N)
        # scaled_y : (B, O, N)
        # x : (B, T, N)
        # y : (B, O, N)
        if isinstance(index, int):
            scaled_x = self.scaled_data[index:index+self.window]

            scaled_y = self.scaled_data[self.window + self.horizon - 1 +
                                 index:self.window + self.horizon - 1 + index+self.steps]
            
            x = self.dataset.data[index:index+self.window]
            
            y = self.dataset.data[self.window + self.horizon - 1 +
                                 index:self.window + self.horizon - 1 + index+self.steps]

            return scaled_x, scaled_y, x , y
        else:
            raise TypeError('Not surpported index type!!!')

    def __len__(self):
        return len(self.dataset) - self.window - self.horizon + 1 - self.steps + 1







class SingleStepWrapper(Dataset):

    def __init__(self, dataset: TimeSeriesDataset, window: int = 168, horizon: int = 3):
        self.dataset = dataset
        self.window = window
        self.horizon = horizon

    def __getitem__(self, index):
        return self.dataset.data[index:index+self.window], self.dataset.data[self.window + self.horizon - 1 + index]

    def __len__(self):
        return len(self.dataset) - self.window - self.horizon + 1


class MultiStepWrapper(Dataset):
    def __init__(self, dataset: TimeSeriesDataset, window: int = 168, horizon: int = 3, steps: int = 2):
        self.dataset = dataset
        self.window = window
        self.horizon = horizon
        self.steps = steps

    def __getitem__(self, index):
        return self.dataset.data[index:index+self.window], self.dataset.data[self.window + self.horizon - 1 + index:self.window + self.horizon - 1 + index+self.steps]

    def __len__(self):
        return len(self.dataset) - self.window - self.horizon + 1 - self.steps + 1


class SingStepFlattenWrapper(Dataset):
    def __init__(self, dataset: TimeSeriesDataset, window: int = 168, horizon: int = 3):
        self.dataset = dataset
        self.window = window
        self.horizon = horizon

    def __getitem__(self, index):
        x = self.dataset.data[index:index+self.window]
        y = self.dataset.data[self.window + self.horizon - 1 + index]
        return x.flatten(), y

    def __len__(self):
        return len(self.dataset) - self.window - self.horizon + 1


class MultiStepFlattenWrapper(Dataset):
    def __init__(self, dataset: TimeSeriesDataset, window: int = 168, horizon: int = 3, steps: int = 2):
        self.dataset = dataset
        self.window = window
        self.horizon = horizon
        self.steps = steps

    def __getitem__(self, index):
        x = self.dataset.data[index:index+self.window]
        y = self.dataset.data[self.window + self.horizon - 1 +
                              index:self.window + self.horizon - 1 + index+self.steps]
        return x.flatten(), y.flatten()

    def __len__(self):
        return len(self.dataset) - self.window - self.horizon + 1 - self.steps + 1



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_timeseries.nn.kernels import MovingAvg
from torch_timeseries.nn.decomp import SeriesDecomp
import numpy as np




class DLinear(nn.Module):
    """
    Decomposition-Linear
    """
    def __init__(self, seq_len, pred_len, enc_in, individual:bool = False, output_prob=0):
        super(DLinear, self).__init__()
        self.seq_len = seq_len
        self.pred_len = pred_len
        self.output_prob = output_prob

        # Decompsition Kernel Size
        kernel_size = 25
        self.decompsition = SeriesDecomp(kernel_size)
        self.individual = individual
        self.channels = enc_in

        if self.individual:
            self.Linear_Seasonal = nn.ModuleList()
            self.Linear_Trend = nn.ModuleList()
            
            for i in range(self.channels):
                self.Linear_Seasonal.append(nn.Linear(self.seq_len,self.pred_len))
                self.Linear_Trend.append(nn.Linear(self.seq_len,self.pred_len))

                # Use this two lines if you want to visualize the weights
                # self.Linear_Seasonal[i].weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
                # self.Linear_Trend[i].weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
        else:
            self.Linear_Seasonal = nn.Linear(self.seq_len,self.pred_len)
            self.Linear_Trend = nn.Linear(self.seq_len,self.pred_len)
            
            # Use this two lines if you want to visualize the weights
            # self.Linear_Seasonal.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
            # self.Linear_Trend.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
        if self.output_prob > 0:
            self.act = F.gelu
            self.projection = nn.Linear(
                enc_in * pred_len, self.output_prob)

    def forward(self, x):
        # x: [Batch, Input length, Channel]
        seasonal_init, trend_init = self.decompsition(x)
        seasonal_init, trend_init = seasonal_init.permute(0,2,1), trend_init.permute(0,2,1)
        if self.individual:
            seasonal_output = torch.zeros([seasonal_init.size(0),seasonal_init.size(1),self.pred_len],dtype=seasonal_init.dtype).to(seasonal_init.device)
            trend_output = torch.zeros([trend_init.size(0),trend_init.size(1),self.pred_len],dtype=trend_init.dtype).to(trend_init.device)
            for i in range(self.channels):
                seasonal_output[:,i,:] = self.Linear_Seasonal[i](seasonal_init[:,i,:])
                trend_output[:,i,:] = self.Linear_Trend[i](trend_init[:,i,:])
        else:
            seasonal_output = self.Linear_Seasonal(seasonal_init)
            trend_output = self.Linear_Trend(trend_init)

        x = seasonal_output + trend_output
        output= x.permute(0,2,1) # to [Batch, Output length, Channel]
        if self.output_prob > 0:
            output = output.reshape(output.shape[0], -1)
            # (batch_size, num_classes)
            output = self.projection(output)
        return output


    # def forecast(self, x_enc):
    #     return self(x_enc)

    # def imputation(self, x_enc):
    #     return self(x_enc)

    # def anomaly_detection(self, x_enc):
    #     return self(x_enc)

    # def classification(self, x_enc):
    #     # Encoder
    #     enc_out = self(x_enc)
    #     # Output
    #     # (batch_size, seq_length * d_model)
    #     output = enc_out.reshape(enc_out.shape[0], -1)
    #     # (batch_size, num_classes)
    #     output = self.projection(output)
    #     return output



import os
import resource
from.dataset import Dataset, Freq, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import download_url, download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np
import torch.utils.data



class ETTm2(TimeSeriesDataset):
    """
    Electricity Transformer Temperature minutely~(15 minutes level). This dataset contains 2-year data from two separate counties in China. Each data point consists of the target value "oil temperature" and 6 power load features.
    
    This dataset is collected by Informer and is available at https://github.com/zhouhaoyi/ETDataset.

    Attributes:
        name (str): Name of the dataset.
        freq (str): Frequency of the data points.
        num_features (int): Number of features in the dataset.
        length (int): Length of the dataset.

    Methods:
        download():
            Downloads and extracts the dataset.
        _load():
            Loads the dataset into a NumPy array.
        _process():
            Processes the dataset using the parent class's process method.
    """
    name:str= 'ETTm2'
    num_features: int = 7
    freq : Freq = 't'
    length : int  = 69680
    
    def download(self):
        download_url(
         "https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTm2.csv",
         self.dir,
         filename='ETTm2.csv',
         md5="7687e47825335860bf58bccb31be0c56"
        )

        
    def _load(self) -> np.ndarray:
        self.file_path = os.path.join(self.dir, 'ETTm2.csv')
        self.df = pd.read_csv(self.file_path,parse_dates=[0])
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.iloc[:, 1:].to_numpy()
        return self.data
    
        


import os
import resource
from.dataset import Dataset, TimeSeriesDataset
from typing import Any, Callable, List, Optional
import torch
from torchvision.datasets.utils import download_url, download_and_extract_archive, check_integrity
import pandas as pd
import numpy as np
import torch.utils.data



class ILI(TimeSeriesDataset):
    """
    ILI (Influenza-Like Illness) records the weekly ratio of influenza-like illness patients versus the total patients 
    by the Centers for Disease Control and Prevention (CDC) of the United States from 2002 to 2021.
    
    This dataset contains 966 data points, each with 7 features, and the data frequency is hourly ('h').

    Collected from: https://gis.cdc.gov/grasp/fluview/fluportaldashboard.html.

    Attributes:
        name (str): Name of the dataset, it is 'ILI'.
        num_features (int): Number of features in the dataset, it is 7.
        length (int): Total length of the dataset, it is 966.
        freq (str): Frequency of the data points, it is 'h'.

    Args:
        TimeSeriesDataset (_type_): The base class for time series datasets.

    Returns:
        np.ndarray: The data loaded from the CSV file.

    Methods:
        download(): Downloads and extracts the dataset archive from the specified URL.
        _load() -> np.ndarray: Loads the data from the CSV file and returns it as a numpy array.
    """
    

    name:str= 'ILI'
    num_features: int = 7
    length : int  = 966
    freq:str = 'h'
    
    def download(self):
        download_and_extract_archive(
         "https://raw.githubusercontent.com/wayne155/multivariate_timeseries_datasets/main/illness.zip",
         self.dir,
         filename='illness.zip',
         md5="e18542fafd509ad52ffedde1b0f1018d"
        )
        
        
    def _load(self) -> np.ndarray:
        self.file_path =os.path.join( os.path.join(self.dir, 'illness')  , 'national_illness.csv')
        self.df = pd.read_csv(self.file_path,parse_dates=[0])
        self.dates = pd.DataFrame({'date':self.df.date})
        self.data = self.df.iloc[:, 1:].to_numpy()
        return self.data
    